package main

import (
	"errors"
	"fmt"
	"git.sr.ht/~geb/numen/vox"
	"git.sr.ht/~geb/opt"
	"io"
	"os"
	"strconv"
	"strings"
	"unicode"
)

var Version string

var usage = `Usage: sprec [OPTIONS] < audio

sprec reads audio from stdin and prints speech recognition results.

--alts NUM          Specify the max number of alternative results.
--partial           Also print intermediate results.
--phrase-file FILE  Limit the vocab to keyphrases in a file.
--version           Print the version and exit.

See 'man sprec' for the documentation.
`

func inform(a ...any) {
	fmt.Fprintln(os.Stderr, "sprec:", fmt.Sprint(a...))
}

func quotesplit(s string) []string {
	var fields []string
	var sb strings.Builder
	quoted := false
	for _, r := range s {
		if r == '"' {
			quoted = !quoted
		} else if !quoted && unicode.IsSpace(r) {
			fields = append(fields, sb.String())
			sb.Reset()
		} else {
			sb.WriteRune(r)
		}
	}
	if sb.Len() > 0 {
		fields = append(fields, sb.String())
	}
	return fields
}

func printResults(results []vox.Result) {
	for i := range results {
		var sb strings.Builder
		var brokenPhrases string
		if len(results[i].Phrases) > 0 {
			n := len(results[i].Phrases) - 1
			for j := range results[i].Phrases {
				n += len(results[i].Phrases[j].Text)
				if j == 0 {
					sb.WriteString(`"` + results[i].Phrases[j].Text + `"`)
				} else {
					sb.WriteString(` "` + results[i].Phrases[j].Text + `"`)
				}
			}
			brokenPhrases = results[i].Text[n:]
		} else {
			brokenPhrases = results[i].Text
		}
		fmt.Printf("final %d: %s\n", i+1, sb.String()+brokenPhrases)
	}
}

func mainish() int {
	var opts struct {
		Alts       int
		Partial    bool
		PhraseFile string
	}

	{
		o := opt.NewOptSet()

		o.Func("alts", func(s string) error {
			var err error
			opts.Alts, err = strconv.Atoi(s)
			return err
		})

		o.FlagFunc("h", func() error {
			fmt.Print(usage)
			os.Exit(0)
			panic("unreachable")
		})
		o.Alias("h", "help")

		o.BoolFunc("partial", func(b bool) error {
			opts.Partial = b
			return nil
		})

		o.Func("phrase-file", func(s string) error {
			opts.PhraseFile = s
			return nil
		})

		o.FlagFunc("version", func() error {
			fmt.Println(Version)
			os.Exit(0)
			panic("unreachable")
		})

		if err := o.Parse(false, os.Args[1:]); err != nil {
			inform(err.Error())
			return 1
		}
		if len(o.Args()) > 0 {
			inform("unexpected argument: " + o.Args()[0])
			return 1
		}
	}

	m := os.Getenv("SPREC_MODEL")
	if m == "" {
		m = "/usr/share/vosk-models/small-en-us"
		if _, err := os.Stat(m); errors.Is(err, os.ErrNotExist) {
			inform("default model needs installed: " + m)
			return 1
		}
	}
	model, err := vox.NewModel(m)
	if err != nil {
		inform(err.Error())
		return 1
	}
	defer model.Free()

	var phrases []string
	if opts.PhraseFile != "" {
		data, err := os.ReadFile(opts.PhraseFile)
		if err != nil {
			inform("failed to read phrases: " + err.Error())
			return 1
		}
		phrases = quotesplit(string(data))
	}

	sampleRate, bitDepth := 16000, 16
	rec, err := vox.NewRecognizer(model, sampleRate, bitDepth, phrases)
	if err != nil {
		inform(err.Error())
		return 1
	}
	defer rec.Free()
	rec.SetMaxAlternatives(opts.Alts)
	rec.SetWords(opts.PhraseFile != "")
	rec.SetKeyphrases(opts.PhraseFile != "")

	for {
		buf := make([]byte, 4096)
		_, err := io.ReadFull(os.Stdin, buf)
		if err != nil {
			if err == io.EOF || err == io.ErrUnexpectedEOF {
				break
			}
			inform(err.Error())
			return 1
		}

		finalized, err := rec.Accept(buf)
		if err != nil {
			panic(err.Error())
		}

		if finalized {
			printResults(rec.FinalResults())
		} else if opts.Partial {
			fmt.Println("partial: " + rec.Results()[0].Text)
		}
	}
	printResults(rec.FinalResults())
	return 0
}

func main() {
	os.Exit(mainish())
}
